<?php
/**
 *  Copyright (c) 2010 angelystor
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */

/**
 * This class stores the audit trail
 * If sequence_number is the same for different rows, it means that those edits were done in a single operation
 * To find out if the record was new, updated, or deleted, look at the difference between new and previous values
 * If prev == null --> new record
 * If prev != null && new != null --> updated record
 * If prev != null && new == null --> deleted record
 *
 * It also stores the affected row id. This is the id of the row that changed, so you can refer to that + the table name, to locate the row in question (unless deleted, of course)
 * @author angelystor
 *
 */
class AuditTrail extends CustomCActiveRecord
{
	/**
	 * The followings are the available columns in table 'audit_trail':
	 * @var double $id
	 * @var string $table_name
	 * @var string $column_name
	 * @var string $previous_value
	 * @var string $new_value
	 * @var string $sys_done_by
	 * @var string $timestamp_date
	 * @var double $sequence_number
	 */
	private static $crud = array('C'=>'Create','U'=>'Update','D'=>'Delete');

	/**
	 * Returns the static model of the specified AR class.
	 * @return CActiveRecord the static model class
	 */
	public static function model($className=__CLASS__)
	{
		return parent::model($className);
	}

	/**
	 * @return string the associated database table name
	 */
	public function tableName()
	{
		return 'audit_trail';
	}

	/**
	 * @return array validation rules for model attributes.
	 */
	public function rules()
	{
		// NOTE: you should only define rules for those attributes that
		// will receive user inputs.
		return array(
			array('sequence_number, affected_row_id', 'numerical'),
			array('table_name, column_name', 'length', 'max'=>255),
			array('type', 'length', 'max'=>500),
			array('previous_value, new_value', 'length', 'max'=>2048),
			array('sys_create_by', 'length', 'max'=>30),
		);
	}

	/**
	 * @return array relational rules.
	 */
	public function relations()
	{
		// NOTE: you may need to adjust the relation name and the related
		// class name for the relations automatically generated below.
		return array(
			'sequence_number' => array(self::BELONGS_TO, 'AuditTrailSequenceNumber', 'sequence_number'),
		);
	}

	/**
	 * @return array customized attribute labels (name=>label)
	 */
	public function attributeLabels()
	{
		return array(
			'id' => 'Id',
			'affected_row_id' => 'Affected Row Id',
			'table_name' => 'Table Name',
			'column_name' => 'Column Name',
			'previous_value' => 'Previous Value',
			'new_value' => 'New Value',
			'sys_create_by' => 'Sys Create By',
			'sys_create_datetime' => 'Timestamp Date',
			'sequence_number' => 'Sequence Number',
		);
	}

	/**
	 * Finds the latest audit entry given the model.
	 * @var $model the model in which we want to find the latest entry for
	 * @return an array of AuditTrail objects which are the latest entries for the model with the same sequence number
	 */
	public static function findLatestAuditEntries($model)
	{
		//by right, this can be just 1 nested query, but not sure how to express this in yii just by adding conditions
		//the query we want is (in pseudosql):
		//select * from audit_trail t where t.sequence_number = (select sequence_number s, max(sys_create_datetime) from audit_trail where table_name=x and affected_row_id=y)
		/*
		$criteria = new CDbCriteria;
		$criteria->addColumnCondition(array('table_name'=>$model->tableName(),
											'affected_row_id'=>$model->id,
									  ));
		$criteria->addCondition('sequence_number = (select max(sequence_number) from '.AuditTrail::model()->tableName());
		*/


		$entries = self::model()->findBySql('SELECT * FROM '.self::model()->tableName().
										  	' WHERE sequence_number = (select max(sequence_number) from '.
											self::model()->tableName().' where table_name="'.$model->tableName().'" AND affected_row_id='.$model->id.')');
		//$criteria->select = array('*, max(sys_create_datetime)', 'sequence_number');

		//$seq_no = self::model()->find($criteria)->sequence_number;
		$seq_no = $entries->sequence_number;

		if (is_null($seq_no)) return array();

		//based on this seq_no, we can get all the records that correspond to this
		$models = self::model()->findAll("sequence_number = $seq_no");
		return $models;
	}

	/**
	 * Returns which action a record is, we only have cud -> create, update, delete
	 * @return the letter C U or D depending on what happened to the record
	 */
	public function whichCRUD()
	{
		//if previous_value is null, this is a new record, C
		if (is_null($this->previous_value)) return 'C';
		//if new_value is null, this is a delete, D
		if (is_null($this->new_value)) return 'D';
		//else we return U, update
		return 'U';
	}
}