<?php
/**
 *  Copyright (c) 2010 angelystor
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */

/**
 * Subclass this class if the required model needs to be backed by an audit trail
 * @author angelystor
 *
 */
class AuditedCActiveRecord extends CustomCActiveRecord
{
	/**
	 * Modify this array to set the columns to skip when saving it to an audit trail
	 * I have a column called sys_update_datetime for all my tables which will be changed everytime a record is saved
	 * I don't want to save this in the audit trail, so it's listed in the array
	 * Values of this array don't matter
	 */
	public $skip_columns = array("sys_update_datetime"=>1,"sys_create_datetime"=>1);

	/**
	 *  this value will be set by the save function, it will contain the sequence number of the batch
	 *  Use this when you want to use the same sequence number for multiple save objects (ie update different tables but
	 *  in the same operation)
	 */
	public $chain_seq_no = null;
	const DELETE = 'D';
	const CREATE = 'C';
	const UPDATE = 'U';

	private $is_changed = false;

	public function save()
	{
		$is_new = $this->getIsNewRecord();

		$old = array();

		if (!$is_new)
		{
			$old = $this->findByPk($this->id)->attributes;
		}

		$saved = parent::save();
		if (!$saved)
		{
			return false;
		}

		if ($is_new)
		{
			$this->chain_seq_no = $this->insertIntoAudit();
		}
		else
		{
			$this->chain_seq_no = $this->saveIntoAudit($old);
		}

		return true;
	}

	public function saveAttributes($attributes)
	{
		$old 		= $this->findByPk($this->id)->attributes;
		$current 	= $this->attributes;
		$saved 		= parent::saveAttributes($attributes);

		if (!$saved)
		{
			return false;
		}

		//before putting it into the audit, we only want the old array to contain attributes which have been inserted
		//while the rest will be taken from current
		$n_old = $current;
		foreach ($attributes as $k=>$v)
		{
			$n_old[$v] = $old[$v];
		}
		$this->chain_seq_no = $this->saveIntoAudit($n_old);
		return true;
	}

	public function delete()
	{
		$this->chain_seq_no = $this->deleteIntoAudit();

		return parent::delete();

	}

	/**
	 * This function creates a new entry into the sequence number table and returns it
	 * @return the audit sequence number to be used
	 */
	public function auditSequenceNumber()
	{
		$audit_seq = new AuditTrailSequenceNumber;
		$audit_seq->transactedSave();
		$seq_no = $audit_seq->id;
		return $seq_no;
	}

	/**
	 * This creates an audit sequence number
	 * @return the audit sequence number
	 */
	public function insertIntoAudit()
	{
		$s_no = $this->saveIntoAudit(array());

		return $s_no;
	}

	/**
	 * Saves changed values into the audit table. This function should be called ONLY for update and delete functions
	 * If $this->seq_no is null, it will create a new sequence number. So if this operation is part of a batch
	 * set this object's seq_no first before calling this method
	 * For insert functions, call insertIntoAudit
	 * @param $old the old data as an array (eg array("col1"=>"value", "col2"=>"value2") )
	 * @return the sequence number (so that if the sequence number is auto, it will return a new one)
	 */
	public function saveIntoAudit($old)
	{
		$columns	= $this->createAuditColumns($this->attributes, $old);
		$seq_no		= $this->saveColumnsIntoAudit($columns, $this->seq_no);

		return $seq_no;

	}

	public function deleteIntoAudit()
	{
		$columns = array();

		foreach ($this->attributes as $k=>$v)
		{
			if (!$this->skipColumns($k))
			{
				$r			= array();
				$r['new']	= null;
				$r['old']	= $v;
				$columns[$k]= $r;
			}
		}

		$seq_no = $this->saveColumnsIntoAudit($columns);
		return $seq_no;
	}

	private function createAuditColumns($input_columns, $old_columns)
	{
		$result = array();

		$this->is_changed = false;
		foreach ($input_columns as $k=>$v)
		{
			if (!$this->skipColumns($k) && !$this->compareOldNewValues($k, $old_columns[$k], $v))
			{
				$r 			= array();
				$r['new'] 	= $v;
				$r['old']	= $old_columns[$k];
				$result[$k]	= $r;

				$this->is_changed = true;
			}
		}
		return $result;
	}

	private function saveColumnsIntoAudit($columns)
	{
		if (is_null($this->seq_no))
		{
			$this->seq_no = $this->auditSequenceNumber();
		}

		foreach ($columns as $k=>$v)
		{
			$audit_trail				= new AuditTrail;
			$audit_trail->table_name	= $this->tableName();
			$audit_trail->type			= get_class($this);
			$audit_trail->affected_row_id=$this->id;
			$audit_trail->column_name	= $k;
			$audit_trail->previous_value= $v['old'];
			$audit_trail->new_value		= $v['new'];
			$audit_trail->sequence_number=$this->seq_no;

			$audit_trail->transactedSave();
		}
		$seq_no = $this->seq_no;
		$this->seq_no = null;

		return $seq_no;
	}


	/**
	 * This method will return true if the column should be skipped, ie not put into the audit table
	 * @return true or false depending on whether the column should be skipped
	 */
	private function skipColumns($column_name)
	{
		return array_key_exists($column_name, $this->skip_columns);
	}

	/**
	 * Semi custom comparator, for most of the values, it will do a simple != check
	 * Modify this method as required. This determines how each new/old value should be checked.
	 * The default is a simple equality check, but refer to the commented out code for the actual check I use
	 *
	 * Description of commented code:
	 * I have an afterFind function that converts dates to human readable data, so for CDbExpressions like now, I need to convert
	 * it to a human readable format so I can compare it properly
	 * Due to this, new/old values are passed in via reference, so this method can modify them.
	 *
	 * NOTE: This method CAN modify the old/new values. This is done mainly for dates (it's ALWAYS about dates), because
	 * afterFind converts it to a human readable date, but we don't necessarily want to store it in the database as such
	 * So it WILL convert to a DB friendly: Y M D format.
	 * @var $column_name the column name. (Only for commented code: This is important because we'll be using endsWith('date') to check denote that it's a date column)
	 * @var $old_value the old value, this MAY be changed.
	 * @var $new_value
	 * @return true if they are the same, false otherwise
	 */
	public function compareOldNewValues($column_name, &$old_value, &$new_value)
	{
		/**
		 * Don't just uncomment this! It requires a utility function to format dates to be database friendly
		 * Date columns for me are assumed to end with date (eg: birth_date, anniversary_date)
		 */
		/*
		if ($this->endsWith($column_name, 'date'))
		{
			if ($old_value != 'now()')
				$old_value = Utility::dateFormatForDatabase($old_value);
			if ($new_value != 'now()')
				$new_value = Utility::dateFormatForDatabase($new_value);
			return $old_value == $new_value;
		}
		else return $old_value == $new_value;
		*/
		return $old_value == $new_value;
	}

	/**
	 * This value is set after the record is saved (it will be determined by the audit trail)
	 * So to find out if a record has been changed, call after saving
	 * @return true if the record has been saved. False otherwise
	 */
	public function isChanged()
	{
		return $this->is_changed;
	}
}
?>