<?php
/**
 *  Copyright (c) 2010 angelystor
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */
/**
 * This class must be the base model class for a project if transactions are used
 * Because save operations are no longer 1 table (due to writing to the audit table) transactions are coded in the class, via transactedXXX methods.
 * However, this feature requires nested transactions (refer to NestedCDbConnection and NestedTransactionPDO)
 * If the database does not support nested transactions, then you should manually begin/end transactions and just call save/delete methods
 *
 * This class also can redefine the find method, which formats date columns to a human readable format. It is commented out here, but you can always uncomment it
 * and add a date format function
 *
 * This class assumes that all tables have sys_create_by, sys_create_datetime, sys_update_by, sys_update_datetime.
 * If they exist, then xxx_create_xxx will be populated when a new record is saved, and if xxx_update_xxx exists, then it will be updated whenever a record is updated
 * These 4 fields are optional
 * @author angelystor
 *
 */
class CustomCActiveRecord extends CActiveRecord
{

	//the sequence number for audit trails, if this is empty or null, the saving process will create a new one
	public $seq_no = null;


	/**
	 * Uncomment to have date formatting for the view
	 */
	/*
	public function init()
	{
		$this->formatAttributes('$this->$k = "";');
	}
	*/
	private function endsWith($fullStr, $endStr)
	{
		$strlen = strlen($endStr);
		$fullStrEnd = substr($fullStr, strlen($fullStr) - $strlen);

		return $fullStrEnd == $endStr;
	}

	/**
	 * Uncomment to have date formatting for the view
	 */
	/*
	private function formatAttributes($function_for_dates)
	{
		$array = $this->attributes;

		foreach ($array as $k=>$v)
		{
			//Yii::trace("After Find $k=>$v");
			if ($this->$k === '')
			{
				$this->$k = "";
			}
			//elseif (strrpos($k, "date") !== false)
			elseif ($this->endsWith($k, "date"))
			{
				//$this->$k = Utility::dateFormatForView($this->$k);
				eval($function_for_dates);
			}
		}
	}

	public function afterFind()
	{
		$this->formatAttributes('$this->$k = Utility::dateFormatForView($this->$k);');
		parent::afterFind();
	}
	*/

	public function beforeSave()
	{
		parent::beforeSave();

		if ($this->isNewRecord)
		{
			echo ("new recodr".get_class($this));
			if (array_key_exists('sys_create_by', $this->attributes))
			{
				$this->sys_create_by = Yii::app()->getUser()->id;
			}
			if (array_key_exists('sys_create_datetime', $this->attributes))
			{
				$this->sys_create_datetime = new CDbExpression('now()');
			}
		}

		$this->updateFields();

		/*
		Yii::trace("--- BEFORE SAVE, PRINTING STUFF ---");
		Yii::trace(print_r($this->getAttributes(), true));
		Yii::trace("--- END BEFORE SAVE, PRINTING STUFF ---");
		*/

		return true;
	}

	private function updateFields()
	{
		if (array_key_exists('sys_update_by', $this->attributes))
		{
			$this->sys_update_by = Yii::app()->getUser()->id;
		}
		if (array_key_exists('sys_update_datetime', $this->attributes))
		{
			$this->sys_update_datetime = new CDbExpression('now()');
		}
	}

	/**
	 * This is kinda a major hack ...
	 * Apparently saveAttributes returns false if there is no change to the record (ie affected row = 0). This will mistakenly cause the program
	 * to think that the save failed, thereby rolling back the transaction (+ do other nefarious things).
	 * We COULD not check the return of this, but I think that's quite important to do... so for now we'll force update_by and update_datetime to be updated even on saveAttributes
	 * Which they should actually, since the record IS being updated.
	 * @see framework/CActiveRecord#saveAttributes($attributes)
	 */
	public function saveAttributes($attributes)
	{
		$this->updateFields();
		if (array_key_exists('sys_update_by', $attributes))
		{
			$attributes[] = 'sys_update_by';
			$attributes[] = 'sys_update_datetime';
		}
		return parent::saveAttributes($attributes);
	}

	public function transactedSave()
	{
		$transaction = $this->dbConnection->beginTransaction();
		try
		{
			if ($this->save())
			{
				$transaction->commit();
				return true;
			}
			else
			{
				$transaction->rollBack();
				return false;
			}
		}
		catch (Exception $e)
		{
			$transaction->rollBack();
			throw $e;
		}
	}

	/**
	 * We don't check saveAttributes' return for success/failure because the return of saveAttributes depends on the number of affected rows
	 * Which CAN be 0 if you're just calling saveAttributes but there is no resultant change in data. This can cause weird failures. So we now depend on
	 * catching the exception. If there is an exception, we will catch it, rollback and rethrow it up for the controller to handle.
	 * @param unknown_type $attributes
	 * @return unknown_type
	 */
	public function transactedSaveAttributes($attributes)
	{
		$transaction = $this->dbConnection->beginTransaction();
		try
		{
			$this->saveAttributes($attributes);
			$transaction->commit();
			return true;
		}
		catch (Exception $e)
		{
			$transaction->rollBack();
			throw $e;
		}
		return false;
	}

	public function transactedDelete()
	{
		$transaction = $this->dbConnection->beginTransaction();
		try
		{
			if ($this->delete())
			{
				$transaction->commit();
				return true;
			}
			else
			{
				$transaction->rollBack();
				return false;
			}
		}
		catch (Exception $e)
		{
			$transaction->rollBack();
			throw $e;
		}

	}

}

?>